/*
    Amstrac CPC Mode 1 specific things
    Mostly based on this: http://www.cpcwiki.eu/index.php/Video_modes
    
    Changelog:
    
    3.5.2016: Overscan exporter
    3.5.2016: Fixed screenbyte, export should work for asm and non-overscan bin
    2.5.2016: The first version
*/

class CPCMode1 extends Machine
{
    // Simple 3x3x3 RGB
    final int rgb[]={#000000,#000080,#0000ff,
                     #800000,#800080,#8000ff,
                     #ff0000,#ff0080,#ff00ff,
                     #008000,#008080,#0080ff,
                     #808000,#808080,#8080ff,
                     #ff8000,#ff8080,#ff80ff,
                     #00ff00,#00ff80,#00ffff,
                     #80ff00,#80ff80,#80ffff,
                     #ffff00,#ffff80,#ffffff};
    
    // Color numbers used by the CPC hardware
    final int hwnumbers[]={0x54,0x44,0x55,
                           0x5c,0x58,0x5d,
                           0x4c,0x45,0x4d,
                           0x56,0x46,0x57,
                           0x5e,0x40,0x5f,
                           0x4e,0x47,0x4f,
                           0x52,0x42,0x53,
                           0x5a,0x59,0x5b,
                           0x4a,0x43,0x4b};
                           
    boolean overscan;
                    
    Button save_bin_b,
           save_asm_b;
    
    CPCMode1(boolean overi)
    {
        overscan=overi;
        number=CPC_MODE1;
        if(overscan)
            name="CPC Mode 1 overscan";
        else
            name="CPC Mode 1";
        
        if(overscan)
        {
            xsize=384;
            ysize=272;
        }
        else
        {
            xsize=320;
            ysize=200;
        }
        blox=gridx=8;
        bloy=gridy=8;
        
        renderx=xsize*prefs.zoom; // Pixels are about 2x as wide as high
        rendery=ysize*prefs.zoom;
        
        colorselheight=(prefs.UIWIDTH-2*prefs.BWIDTH)*3/2;
        
        border=1;
        maxborder=26;
        pen=1;
        
        errorcolor=#dc9a65;
        gridcolor=#405040;
        
        super.rgb=rgb;
    }
    
    // Draw the color selector
    void drawcolorsel(int px,int py)
    {
        int wide=(prefs.UIWIDTH-2*prefs.BWIDTH)/3,
            high=wide/2,
            selectedx=0,
            selectedy=0,
            selected=0;
        
        for(int y=0,i=0;y<rgb.length/3;y++)
        {
            for(int x=0;x<3;x++,i++)
            {
                fill(rgb[i]);
                rect(px+x*wide,py+y*high, wide,high);
                
                if(mousePressed) // A bit of a hack, yes
                {
                    if(inside(px+x*wide,py+y*high, px+x*wide+wide,py+y*high+high))
                        if(i<=maxborder)
                            border=i;
                }

                if(i==border)
                {
                    selected=i;
                    selectedx=px+x*wide;
                    selectedy=py+y*high;
                }
            }
        }
        
        noFill();
        strokeWeight(2);
        stroke(rgb[selected]^0x808080);
        rect(selectedx,selectedy, wide,high);
        strokeWeight(1);
        noStroke();
        
        fill(prefs.UITEXT);
        text("border: "+str(border),px,py+high*rgb.length/3+prefs.UIROW-7);
    }
    
    // Simple checker: find the optimal 16 color palette
    boolean check(boolean fix)
    {
        boolean errorfound=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        pic.loadPixels();
        
        int count[][]=countcolors(pic.pixels,rgb);
        
        if(count.length>4)
        {
            for(int i=0;i<xsize*ysize;i++)
            {
                boolean colorok=false;
                for(int j=0;j<4;j++)
                    if(count[j][0]==bestmatch(pic.pixels[i],rgb))
                        colorok=true;
                
                if(!colorok) // Wasn't one of the 16
                {
                    if(fix)
                    {
                        int diff=1000000,idx=0;
                        for(int j=0;j<4;j++)
                            if(rgbdist(count[j][1],pic.pixels[i])<diff)
                            {
                                diff=rgbdist(count[j][1],pic.pixels[i]);
                                idx=j;
                            }
                        pic.pixels[i]=count[idx][1];
                    }
                    else
                    {
                        pic.pixels[i]=errorcolor;
                    }
                }
            }
            pic.updatePixels();
            
            message("Too many colors: "+str(count.length));
            return false;
        }
        else
        {
            message("Colors ok");
            return true;
        }
    }

    // Create the palette and bitmap
    void populate()
    {
        colormap=new int[4];
        if(overscan)
            bitmap=new int[1536+2048*7+1728+2048*7];
        else
            bitmap=new int[16384];
        
        pic.loadPixels();
        
        int count[][]=countcolors(pic.pixels,rgb);
        
        if(count.length>4)
        {
            message("Unchecked image!");
            return;
        }
        
        for(int i=0;i<count.length;i++) // Palette done
            colormap[i]=count[i][0];
            
        for(int i=0;i<xsize*ysize;i+=4)
        {
            int a=0,b=0,c=0,d=0;
            for(int j=0;j<count.length;j++)
            {
                if(pic.pixels[i]==count[j][1])
                    a=j;
                if(pic.pixels[i+1]==count[j][1])
                    b=j;
                if(pic.pixels[i+2]==count[j][1])
                    c=j;
                if(pic.pixels[i+3]==count[j][1])
                    d=j;
            }
            
            // Of course the pixels aren't there after one another
            bitmap[screenoffset(i)]=screenbyte(a,b,c,d);
        }
    }
    
    // Make a screen byte out of four colors
    int screenbyte(int a,int b,int c,int d)
    {
        int e=0;
        
        if((a&2)>0) e|=8;
        if((a&1)>0) e|=0x80;
        
        if((b&2)>0) e|=4;
        if((b&1)>0) e|=0x40;
        
        if((c&2)>0) e|=2;
        if((c&1)>0) e|=0x20;
        
        if((d&2)>0) e|=1;
        if((d&1)>0) e|=0x10;
        
        return e;
    }
    // Make a CPC screen offset out of pixel offsets
    int screenoffset(int off)
    {
        int y=off/xsize;  
        if(overscan)
        {
            int val=off%xsize/4 + y%8*2048 + y/8*(xsize/4);
            if(y/8*(xsize/4)>=1536)
                val+=16384-2048;
            return val;
        }
        else
            return off%xsize/4 + y%8*2048 + y/8*(xsize/4);
    }
    
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_bin_b=new Button(x,super.addsavebuttons(x,y),"save bin");
        save_asm_b=new Button(x,save_bin_b.next,"save asm");
        return save_asm_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_bin_b.mouseover() || save_asm_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_bin_b.mouseover()) // BIN
                    {
                        String s=fileselector(prefs.path,"BIN files",new String[] {"bin"},SAVE);
                        if(s!=null)
                            save_bin(s);
                    }
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has too many colors");
                }
            }
        }
    }
    
    // Save BIN
    void save_bin(String name)
    {
        byte b[];
        
        if(pic==null)
            return;
            
        populate(); // Create bitmap/palette

        if(overscan)
        {
            // Read template
            b=loadBytes("templates/cpc-mode1overscan.bin");
            
            // Replace some bytes
            b[85]=(byte)border;
            
            for(int i=0;i<bitmap.length;i++)
                b[256+i]=(byte)bitmap[i];
            for(int i=0;i<colormap.length;i++)
                b[79+i]=(byte)colormap[i];
        }
        else
        {
            // Read template
            b=loadBytes("templates/cpc-mode1.bin");
            
            // Replace some bytes
            b[16457]=(byte)border;
            
            for(int i=0;i<bitmap.length;i++)
                b[69+i]=(byte)bitmap[i];
            for(int i=0;i<colormap.length;i++)
                b[16453+i]=(byte)colormap[i];
        }
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap
        
        PrintWriter f=createWriter(name);
        
        if(overscan)
            f.println("; This needs to be loaded at $200");
        f.print("bmap");
        printarr(f,bitmap,"db","$");
        f.println();
        f.println("; BIOS and hw color numbers");
        f.print("cmap");
        printarr(f,colormap,"db","$");
        int hwcolor[]=new int[4];
        for(int i=0;i<4;i++)
            hwcolor[i]=hwnumbers[colormap[i]];
        f.print("hwmap");
        printarr(f,hwcolor,"db","$");
        f.println();
        f.println("border\tdb\t$"+hex(border,2));
        f.println("hwbord\tdb\t$"+hex(hwnumbers[border],2));

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
