/*
    MSX Screen 3 specific things
    
    Changelog:
    
    20.5.2016: Exporters and importers all done
    20.5.2016: The first version. This mode is supported more for the sake of completeness than real use.
*/

class MsxScreen3 extends Machine
{
    // Supposedly from the TI datasheet
    final int rgb[]={ILLEGAL,#000000,#21c842,#5edc78,
                     #5455ed,#7d76fc,#d4524d,#42ebf5,
                     #fc5554,#ff7978,#d4c154,#e6ce80,
                     #21b03b,#c95bba,#cccccc,#ffffff};
                     
    Button save_com_b,
           load_sc3_b,
           save_sc3_b,
           save_asm_b;
    
    MsxScreen3()
    {
        number=MSX_SCREEN3;
        name="MSX Screen 3";

        xsize=64;
        ysize=48;
        blox=bloy=gridx=gridy=1;
        
        renderx=xsize*prefs.zoom*4;
        rendery=ysize*prefs.zoom*4;
        if(prefs.zoom>1)
        {
            if(prefs.aspect==PAL)
                renderx=xsize*prefs.zoom*12/10*4; // Approximate stretch (it IS flat!)
        }
       
        colorselheight=prefs.UIWIDTH-2*prefs.BWIDTH;
       
        border=4;
        maxborder=15;
        pen=4;
        errorcolor=#ffff33;
        gridcolor=#778822;

        super.rgb=rgb;
    }
    
    // There can't be clashes
    boolean check(boolean fix)
    {
        message("Colors ok (this is a free pixel mode)");
        return true;
    }
    
    // Create a bitmap and the color maps
    void populate()
    {
        colormap=new int[64*48/2];
        
        pic.loadPixels();
        for(int y=0;y<ysize;y++) // Walk through chars
            for(int x=0;x<xsize;x+=2)
            {
                int col1=bestmatch(pic.pixels[y*xsize+x],rgb),
                    col2=bestmatch(pic.pixels[y*xsize+x+1],rgb);
                    
                colormap[x/2*8 + y%8 + y/8*32*8]=(col1<<4)+col2;
            }
    }
    
    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
        load_sc3_b=new Button(x,super.addloadbuttons(x,y),"load sc3");
        return load_sc3_b.next;
    }    
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_com_b=new Button(x,super.addsavebuttons(x,y),"save com");
        save_sc3_b=new Button(x,save_com_b.next,"save sc3");
        save_asm_b=new Button(x,save_sc3_b.next,"save asm");
        return save_asm_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_com_b.mouseover() || save_sc3_b.mouseover() || save_asm_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                if(save_com_b.mouseover()) // COM
                {
                    String s=fileselector(prefs.path,"COM files",new String[] {"com"},SAVE);
                    if(s!=null)
                        save_com(s);
                }
                if(save_sc3_b.mouseover()) // SC3
                {
                    String s=fileselector(prefs.path,"SC3 files",new String[] {"sc3"},SAVE);
                    if(s!=null)
                        save_sc3(s);
                }
                if(save_asm_b.mouseover()) // Asm
                {
                    String s=fileselector(prefs.path,"Asm files",new String[] {"s","asm"},SAVE);
                    if(s!=null)
                        save_asm(s);
                }
            }
        }
        
        if(load_sc3_b.mouseover())
        {
            String s=fileselector(prefs.path,"SC3 files",new String[] {"sc3"},LOAD);
            if(s!=null)
                load_sc3(s);
        }
    }
    
    // Save COM
    void save_com(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/msx-screen3.com");
        
        populate(); // Create the colormap  
        
        // Replace some bytes
        b[1607]=(byte)border;
        for(int i=0;i<colormap.length;i++)
            b[71+i]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Save SC3
    void save_sc3(String name)
    {
        int header[]={0xfe,0,0,0xff,0x5,0,0},
            off=0;
        
        if(pic==null)
            return;

        byte b[]=new byte[32*48+7];
        
        populate(); // Create the bitmap/colormap
        
        for(int i=0;i<header.length;i++) // Header
            b[off++]=(byte)header[i];
        for(int i=0;i<colormap.length;i++) // Colormap
            b[off++]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    // Load SC3
    void load_sc3(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<1543)
        {
            message("Can't open file: "+name);
            return;
        }
        message("Loaded: "+name);
        
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        for(int y=0;y<ysize;y++)
        {
            for(int x=0;x<xsize;x+=2)
            {
                int off=x/2*8 + y%8 + y/8*32*8 +7,
                    col1=((int)b[off])&0xff,
                    col2;
                    
                col2=col1&0xf;
                col1>>=4;
                    
                if(col1==0) col1=1; // We don't want border colors here
                if(col2==0) col2=1;
                
                sourcepic.pixels[y*xsize+x]=rgb[col1];
                sourcepic.pixels[y*xsize+x+1]=rgb[col2];
            }
        }
        
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }

    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap
        
        PrintWriter f=createWriter(name);
        
        f.print("colors:");
        printarr(f,colormap,"db","#");
        f.println();
        f.println("border:\tdb\t#"+hex(border,2));  

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
