/*
    Sharp MZ-700 specific things
    
    Changelog:
    
    22.5.2016: Smarter pseudographics, 2nd charset (Euro) support
    21.5.2016: Tracer for non-blocky chars
    21.5.2016: MZF exporter
    21.5.2016: check() and populate() for pseudographics, asm exporter
    21.5.2016: First version
*/

class SharpMZ700 extends Machine
{
    // Simple 8-bit RGB
    final int rgb[]={#000000,#0000FF,#FF0000,#FF00FF,
                     #00FF00,#00FFFF,#FFFF00,#FFFFFF};
    
    int xorcount[],
        charmap[],
        charbits[][];
    
    boolean pseudo,
            euro;
                     
    Button save_mzf_b,
           save_asm_b;
           
    Checkbutton pseudo_b,
                euro_b;
    
    SharpMZ700()
    {
        number=SHARP_MZ700;
        name="Sharp MZ-700";
        
        xsize=320;
        ysize=200;
        blox=bloy=gridx=gridy=8;
       
        renderx=xsize*prefs.zoom;
        if(prefs.aspect==PAL)
            rendery=ysize*prefs.zoom*5/4; // They're higher than wide, but this is mostly a guess
        else
            rendery=ysize*prefs.zoom;

        colorselheight=(prefs.UIWIDTH-2*prefs.BWIDTH)/2;
        
        border=0;
        maxborder=0;
        pen=1;
        
        super.rgb=rgb;
        pseudo=euro=false;
        
        // Precalc for samebits()
        xorcount=new int[256];
        for(int i=0;i<256;i++)
        {
            int n=0;
            for(int j=0;j<8;j++)
                if((i&(0x80>>j))!=0)
                    n++;
                
            xorcount[i]=8-n;
        }
        
        // Load ROM font and turn it into bits
        PImage p=loadImage("mz-700-char.png");
        p.loadPixels();
        charbits=new int[512][8];
        
        for(int i=0;i<512;i++)
        {
            for(int y=0;y<8;y++)
            {
                int b=0;
                for(int x=0;x<8;x++)
                    if((p.pixels[i*64+y*8+x]&2) != 0)
                        b+=0x80>>x;
                
                charbits[i][y]=b;
            }
        }
    } 
    
    // Color clash + char pixel checker
    boolean check(boolean fix)
    {
        boolean errorfound,badblocks=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        // The superclass checks clashes already
        errorfound=!super.check(fix);
        if(errorfound && !fix)
            return false;
                
        pic.loadPixels();
        
        if(pseudo) // Blocky blocks
        {
            for(int y=0;y<ysize;y+=bloy)
            {
                for(int x=0;x<xsize;x+=blox)
                {
                    int pix[]=getpixels(x,y, blox,bloy);
                    int count[][]=countcolors(pix,rgb);
                    
                    if(count.length>1)
                    {
                        // Check the blocky pixels
                        for(int i=0;i<bloy;i+=bloy/2)
                            for(int j=0;j<blox;j+=blox/2)
                            {
                                int count2[][]=countcolors(getpixels(x+j,y+i, blox/2,bloy/2),rgb);
                                
                                if(count2.length>1) // Too many
                                {
                                    badblocks=true;
                                    if(fix)
                                        setpixels(x+j,y+i, blox/2,bloy/2, count2[0][1]);
                                    else
                                        setpixels(x,y,blox,bloy,errorcolor);
                                }
                            }
                    }
                }
            }
        }
        else // Try to estimate characters
        {
            for(int y=0;y<ysize;y+=bloy)
            {
                for(int x=0;x<xsize;x+=blox)
                {
                    int pix[]=getpixels(x,y, blox,bloy),
                        count[][]=countcolors(pix,rgb),
                        bits[]=bits(pix);
                     
                    boolean found=false;
                    
                    int best=0,
                        diff=-1;
                    
                    for(int i=0;i<((euro)?512:256);i++)
                    {
                        int result=samebits(bits,charbits[i]);
                        
                        if(result==0 || result==64) // Perfect match
                        {
                            found=true;
                            break;
                        }
                        
                        if(result<32)
                            result=64-result;
                            
                        if(result>diff) // Better but not perfect match?
                        {
                            diff=result;
                            best=i;
                        }
                    }
                    
                    if(!found) // No perfect match for this block
                    {
                        badblocks=true;
                        if(fix)
                        {
                            boolean invert=false;
                            
                            if(samebits(charbits[best],bits)<32)
                                invert=true;
                            
                            for(int i=0;i<bloy;i++)
                                for(int j=0;j<blox;j++)
                                    if((charbits[best][i]&(0x80>>j))!=0)
                                    {
                                        if(invert)
                                            pix[i*blox+j]=count[0][1];
                                        else
                                            pix[i*blox+j]=count[1][1];
                                    }
                                    else
                                    {
                                        if(invert)
                                            pix[i*blox+j]=count[1][1];
                                        else
                                            pix[i*blox+j]=count[0][1];
                                    }
                                        
                            setpixels(x,y, blox,bloy, pix);
                        }
                        else
                        {
                            setpixels(x,y, blox,bloy, errorcolor);
                        }
                    }
                }
            }
        }
        
        pic.updatePixels();
        
        if(badblocks)
        {
            if(errorfound)
                message("Color clashes and illegal blocks found");
            else
                message("Illegal blocks found");
            return false;
        }
        else
        {
            if(errorfound)
                return false;
        }
        message("Colors and blocks ok");
        return true;
    }
    
    // Allow 320x200 pixel images and 80x50 pseudographics
    boolean correctsize(int x,int y)
    {
        if((x==xsize && y==ysize) || (x==80 && y==50))
            return true;
        else
            return false;
    }
    
    // Find number of same bits
    int samebits(int a[],int b[])
    {
        int same=0;
        for(int i=0;i<a.length;i++)
            same+=xorcount[a[i]^b[i]];
        
        return same;
    }
    // Return the bits: most common color bg (0), others fg (1)
    int[] bits(int a[])
    {
        int count[][]=countcolors(a,rgb),
            bits[]=new int[8];

        // Count all            
        for(int i=0;i<8;i++)
            for(int j=0;j<8;j++)
                if(a[i*8+j]!=count[0][1])
                    bits[i]+=(0x80>>j);
        
        return bits;
    }
    
    // Create chars/colormap in MZ-700 format. Needs to be checked before.
    void populate()
    {
        // Char nums for pseudogfx. +0x100 = invert colors.
        final int lookup[]={0xf0, 0xf1, 0xf2, 0x7a,
                            0xf4, 0x7b, 0x6c, 0x1f8,
                            0xf8, 0x5b, 0x3b, 0x1f4,
                            0x3a, 0x1f2, 0xf1, 0x43};
        
        colormap=new int[(xsize/blox)*(ysize/blox)];
        charmap=new int[(xsize/blox)*(ysize/blox)];
        
        pic.loadPixels();
        
        if(pseudo) // Blocks
        {
            for(int y=0,i=0;y<ysize;y+=bloy)
            {
                for(int x=0;x<xsize;x+=blox,i++)
                {
                    int pix[]=getpixels(x,y, blox,bloy),
                        count[][]=countcolors(pix,rgb),
                        bits[]=bits(pix);

                    // Char out of pix
                    int cnum=0;
                    if((bits[0]&0x80)!=0)
                        cnum+=1;
                    if((bits[0]&0x8)!=0)
                        cnum+=2;
                    if((bits[4]&0x80)!=0)
                        cnum+=4;
                    if((bits[4]&0x8)!=0)
                        cnum+=8;
                    charmap[i]=lookup[cnum]&0xff;

                    colormap[i]=count[0][0]; // Colors found
                    if(count.length>1)
                    {
                        if((lookup[cnum]&0x100)!=0) // Invert colors
                        {
                            colormap[i]=count[1][0];
                            colormap[i]+=count[0][0]<<4;
                        }
                        else
                            colormap[i]+=count[1][0]<<4;
                    }
                }
            }
        }
        else // Characters
        {
            for(int y=0,i=0;y<ysize;y+=bloy)
            {
                for(int x=0;x<xsize;x+=blox,i++)
                {
                    int pix[]=getpixels(x,y, blox,bloy),
                        count[][]=countcolors(pix,rgb),
                        bits[]=bits(pix);
                    
                    int j,result=0;
                    for(j=0;j<((euro)?512:256);j++) // Find the correct match
                    {
                        result=samebits(bits,charbits[j]);
                        
                        if(result==0 || result==64) // Perfect match
                            break;
                    }
                    
                    // Char and color
                    charmap[i]=j;
                    if(result==0) // Invert
                    {
                        colormap[i]+=count[0][0]<<4;
                        if(count.length>1)
                            colormap[i]+=count[1][0];
                    }
                    else
                    {
                        colormap[i]+=count[0][0];
                        if(count.length>1)
                            colormap[i]+=count[1][0]<<4;
                    }
                    
                    // 2nd charset
                    if(j>255)
                        colormap[i]|=0x80;
                }
            }
        }
    }

    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_mzf_b=new Button(x,super.addsavebuttons(x,y),"save mzf");
        save_asm_b=new Button(x,save_mzf_b.next,"save asm");
        return save_asm_b.next;
    }
    int addextrabuttons(int x,int y)
    {
        pseudo_b=new Checkbutton(x,super.addextrabuttons(x,y)+prefs.UIGAP,"pseudogfx");
        pseudo_b.checked=pseudo;
        euro_b=new Checkbutton(x,pseudo_b.next,"2nd charset");
        euro_b.checked=euro;
        return euro_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_asm_b.mouseover() || save_mzf_b.mouseover())
        {
            if(pic==null)
                message("Can't save an empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                    if(save_mzf_b.mouseover()) // MZF
                    {
                        String s=fileselector(prefs.path,"MZF files",new String[] {"mzf"},SAVE);
                        if(s!=null)
                            save_mzf(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
        
        if(pseudo_b.mouseover())
        {
            pseudo=!pseudo;
            pseudo_b.checked=pseudo;
            colorcorrect();
        }
        if(euro_b.mouseover())
        {
            euro=!euro;
            euro_b.checked=euro;
            colorcorrect();
        }
    }

    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the charmap/colormap
       
        PrintWriter f=createWriter(name);
        
        f.print("chars");
        printarr(f,charmap,"db","$");
        f.println();
        f.print("colors");
        printarr(f,colormap,"db","$");
        f.println();

        f.flush();
        f.close();
        message("Saved: "+name);
    }
    
    // Save MZF emulator tape
    void save_mzf(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/mz-700.mzf");
        
        populate(); // Create charmap/colormap
        
        for(int i=0;i<charmap.length;i++)
            b[231+i]=(byte)charmap[i];
        for(int i=0;i<colormap.length;i++)
            b[1231+i]=(byte)colormap[i];
        
        saveBytes(name,b);
        message("Saved: "+name);
    }
}
