/*
    Sharp MZ-800 specific things (mode 0, 320x200 4 or 16 colors)
    
    Changelog:

    24.5.2016: New template in use (16k), proper mode names
    23.5.2016: More faithful palette
    23.5.2016: 32k exporter too
    22.5.2016: check(), populate(), asm/MZF export implemented for 16k
    22.5.2016: First version
*/

class SharpMZ800 extends Machine
{
    // Borrowed from mz800emu by Chaky
    final int rgb[]={#000000,#4040ac,#d03400,#b40c8c,
                     #406c00,#24ccff,#e8d430,#d0d0d0,
                     #848484,#008ce8,#ff0000,#f054cc,
                     #54ff54,#80ffff,#ffff28,#ffffff};
    
    int bitmaps[][],
        palette[];
    
    boolean expanded;
                     
    Button save_mzf_b,
           save_asm_b;
           
    Checkbutton expanded_b;
    
    SharpMZ800()
    {
        number=SHARP_MZ800;
        name="Sharp MZ-800 Mode 1/2";
        
        xsize=320;
        ysize=200;
        blox=bloy=gridx=gridy=8;
       
        renderx=xsize*prefs.zoom;
        if(prefs.aspect==PAL)
            rendery=ysize*prefs.zoom*5/4; // They're higher than wide, but this is mostly a guess
        else
            rendery=ysize*prefs.zoom;

        colorselheight=prefs.UIWIDTH-2*prefs.BWIDTH;
        
        border=0;
        maxborder=15;
        pen=9;
        
        expanded=false;
        super.rgb=rgb;
    } 
    
    // Check number of colors
    boolean check(boolean fix)
    {
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        // 16 colors, fine
        if(expanded)
        {
            message("Colors ok");
            return true;
        }
                
        // Drop to 4
        pic.loadPixels();
        int count[][]=countcolors(pic.pixels,rgb);
        
        if(count.length>4)
        {
            for(int i=0;i<xsize*ysize;i++)
            {
                boolean colorok=false;
                for(int j=0;j<4;j++)
                    if(count[j][0]==bestmatch(pic.pixels[i],rgb))
                        colorok=true;
                
                if(!colorok) // Wasn't one of the 16
                {
                    if(fix)
                    {
                        int diff=1000000,idx=0;
                        for(int j=0;j<4;j++)
                            if(rgbdist(count[j][1],pic.pixels[i])<diff)
                            {
                                diff=rgbdist(count[j][1],pic.pixels[i]);
                                idx=j;
                            }
                        pic.pixels[i]=count[idx][1];
                    }
                    else
                    {
                        pic.pixels[i]=errorcolor;
                    }
                }
            }
            
            pic.updatePixels();
            message("Too many colors: "+str(count.length));
            return false;
        }
        else
        {
            message("Colors ok");
            return true;
        }
    }
    
    // Create the bitmaps
    void populate()
    {
        if(expanded)
        {
            palette=new int[16];
            for(int i=0;i<16;i++)
                palette[i]=i;
                
            bitmaps=new int[4][8000];
        }
        else
        {
            palette=new int[4];
            bitmaps=new int[2][8000];
        }
        
        pic.loadPixels();
        
        int count[][]=countcolors(pic.pixels,rgb);
        
        if(!expanded && count.length>4)
        {
            message("Unchecked image!");
            return;
        }
        
        if(!expanded)
            for(int i=0;i<count.length;i++) // RGB palette
                palette[i]=count[i][1];
        
        // Bitmaps then
        for(int i=0,boff=0;i<xsize*ysize;i+=8,boff++)
        {
            for(int j=0;j<8;j++)
            {
                int col;
                
                if(expanded)
                    col=bestmatch(pic.pixels[i+j],rgb);
                else
                    col=bestmatch(pic.pixels[i+j],palette);
                
                if((col&1)!=0) bitmaps[0][boff]+=1<<j;
                if((col&2)!=0) bitmaps[1][boff]+=1<<j;
                if((col&4)!=0) bitmaps[2][boff]+=1<<j;
                if((col&8)!=0) bitmaps[3][boff]+=1<<j;
            }
        }
        
        if(!expanded)
            for(int i=0;i<count.length;i++) // Colornum palette
                palette[i]=count[i][0];
    }

    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_mzf_b=new Button(x,super.addsavebuttons(x,y),"save mzf");
        save_asm_b=new Button(x,save_mzf_b.next,"save asm");
        return save_asm_b.next;
    }
    int addextrabuttons(int x,int y)
    {
        expanded_b=new Checkbutton(x,super.addextrabuttons(x,y),"32k vram");
        expanded_b.checked=expanded;
        return expanded_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_asm_b.mouseover() || save_mzf_b.mouseover())
        {
            if(pic==null)
                message("Can't save an empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                    if(save_mzf_b.mouseover()) // MZF
                    {
                        String s=fileselector(prefs.path,"MZF files",new String[] {"mzf"},SAVE);
                        if(s!=null)
                            save_mzf(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
        
        if(expanded_b.mouseover())
        {
            expanded=!expanded;
            expanded_b.checked=expanded;
            colorcorrect();
        }
    }

    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the charmap/colormap
       
        PrintWriter f=createWriter(name);
        
        f.print("palette");
        printarr(f,palette,"db","$");
        f.println();
        f.println("border\tdb\t$"+hex(border,2));
        f.println();
        for(int i=0;i<((expanded)?4:2);i++)
        {
            f.print("bitmap"+str(i));
            printarr(f,bitmaps[i],"db","$");
            f.println();
        }

        f.flush();
        f.close();
        message("Saved: "+name);
    }
    
    // Save MZF emulator tape
    void save_mzf(String name)
    {
        byte b[];
        
        if(pic==null)
            return;
            
        populate(); // Create bitmaps & palette

        if(expanded)
        {
            // Read template
            b=loadBytes("templates/mz-800-32k.mzf");
            
            b[525]=(byte)border;
            for(int i=0;i<4;i++)
                for(int j=0;j<8000;j++)
                    b[526+i*8000+j]=(byte)bitmaps[i][j];
        }
        else
        {
            // Read template
            b=loadBytes("templates/mz-800.mzf");
            
            b[353]=(byte)border;
            for(int i=0;i<palette.length;i++)
                b[349+i]=(byte)palette[i];
            for(int i=0;i<2;i++)
                for(int j=0;j<8000;j++)
                    b[354+i*8000+j]=(byte)bitmaps[i][j];
        }
        
        saveBytes(name,b);
        message("Saved: "+name);
    }
}
