/*
    Oric specific things
    
    Changelog:
    
    24.5.2016: HRS reading and writing
    14.5.2016: More optimization: do not even check suitability if out ink/paper combo is covered
    14.5.2016: Remove duplicate unnecessary output combos
    13.5.2016: Preliminary forcing in place
    12.5.2016: Allow inverse+ink switch too, some optimization needed soon
    12.5.2016: Oh. Attribute changes can have the invert bit on too.
    8.5.2016: Fixed my own bug with color count testing
    8.5.2016: TAP export
    8.5.2016: The checker should be faster and working now, asm export done
    6.5.2016: Finally a mostly working checker - wasn't easy and there's still some glitch
    4.5.2016: The first version
*/

class Oric extends Machine
{
    // Simple 8-bit RGB
    final int rgb[]={#000000,#ff0000,#00ff00,#ffff00,
                     #0000ff,#ff00ff,#00ffff,#ffffff};
    
    int screen[];
    
    Button load_hrs_b;
    
    Button save_asm_b,
           save_tap_b,
           save_hrs_b;
    
    Oric()
    {
        number=ORIC;
        name="Oric Hires";
        
        xsize=240;
        ysize=200;
        blox=gridx=6;
        bloy=8; // Not really used
        gridy=200;

        // They're not quite square        
        renderx=xsize*prefs.zoom;
        rendery=ysize*prefs.zoom;
        if(prefs.zoom>1)
            renderx=renderx*5/4;

        colorselheight=(prefs.UIWIDTH-2*prefs.BWIDTH)/2;
        
        border=0;
        maxborder=7;
        pen=7;
        errorcolor=#808080;
        gridcolor=#808080;
        
        super.rgb=rgb;
    }

    int blocks[][]; // Blocks and their colors

    // Need to account for the brightness
    boolean check(boolean fix)
    {
        boolean errorfound=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        blocks=new int[xsize/blox][0];
        screen=new int[xsize/blox*ysize];
        pic.loadPixels();

        // Collect the blocks
        for(int y=0;y<ysize;y++)
        {
            for(int x=0;x<xsize;x+=blox)
            {
                int pix[]=getpixels(x,y,blox,1);
                int count[][]=countcolors(pix,rgb);

                blocks[x/blox]=new int[count.length];
                for(int i=0;i<count.length;i++)
                    blocks[x/blox][i]=count[i][0];
            }
            
            // Do the heavy crunching
            ArrayList<int[]> arr[]=new ArrayList[40]; // An arraylist for all the columns
            for(int i=0;i<40;i++)
                arr[i]=new ArrayList();
               
            // Build color combinations to arraylists 
            for(int i=0;i<40;i++)
            {
                if(blocks[i].length==1) // Single color
                {
                    int col=blocks[i][0];
                    
                    // Change paper to this right now, retain ink
                    for(int j=0;j<8;j++)
                    {
                        for(int k=0;k<8;k++)
                        {
                            addblock(arr[i], j,k, 0x10+col,     j,col);
                            addblock(arr[i], j,k, 0x90+(col^7), j,col^7); // inverse
                        }
                    }
                    
                    // Change ink, draw with previous bg
                    for(int j=0;j<8;j++)
                    {
                        for(int k=0;k<8;k++)
                        {
                            addblock(arr[i], j,col,   k,      k,col);
                            addblock(arr[i], j,col^7, 0x80+k, k,col^7);
                        }
                    }
                    
                    // Draw pixels with the previous ink color
                    for(int j=0;j<8;j++)
                    {
                        addblock(arr[i], col,j,   0x40+0x3f, col,j);
                        addblock(arr[i], col^7,j, 0xc0+0x3f, col^7,j); // inverse
                    }
                        
                    // Draw pixels with the previous paper color
                    for(int j=0;j<8;j++)
                    {
                        addblock(arr[i], j,col,   0x40, j,col);
                        addblock(arr[i], j,col^7, 0xc0, j,col^7); // inverse
                    }
                }
                else // Two or more
                {
                    if(blocks[i].length>2)
                    {
                        errorfound=true;
                        if(!fix)
                            break;
                    }
                    
                    int pix[]=getpixels(i*blox,y,blox,1),
                        pix2[]=getpixels(i*blox,y,blox,1);
                        
                    int col1=blocks[i][0],
                        col2=blocks[i][1];
                    
                    for(int j=0;j<pix.length;j++) // Force to 2 colors if needed
                    {
                        pix[j]=bestmatch(pix[j],rgb);
                        if(pix[j]!=col1 && pix[j]!=col2)
                        {
                            if(rgbdist(rgb[pix[j]],rgb[col1])<rgbdist(rgb[pix[j]],rgb[col2]))
                                pix[j]=col1;
                            else
                                pix[j]=col2;
                        }
                        pix2[j]=rgb[pix[j]];
                    }
                    
                    setpixels(i*blox,y,blox,1,pix2);
                        
                    // These in (and out) pairs acceptable
                    addblock(arr[i], col1,col2,     0x40+bits(pix,col1), col1,col2);
                    addblock(arr[i], col2,col1,     0x40+bits(pix,col2), col2,col1);
                    addblock(arr[i], col1^7,col2^7, 0xc0+bits(pix,col1), col1^7,col2^7); // inverse
                    addblock(arr[i], col2^7,col1^7, 0xc0+bits(pix,col2), col2^7,col1^7);
                }
            }

            // Start eliminating      
            for(int i=0;i<40;i++) // Each column
            {
                boolean reserved[][]=new boolean[8][8];
                
                for(int j=0;j<arr[i].size();j++) // All my cobinations
                {
                    int row[]=arr[i].get(j);
                    if(i==0) // First: 7,0
                    {
                        if(row[0]!=7 || row[1]!=0)
                        {
                            arr[i].remove(j);
                            j--;
                        }
                    }
                    else
                    {
                        boolean found=false;
                        
                        if(!reserved[row[3]][row[4]]) // No use duplicating out ink/paper combos
                        {
                            for(int k=0;k<arr[i-1].size();k++) // All previous combinations
                            {
                                int prev[]=arr[i-1].get(k);
                                if(row[0]==prev[3] && row[1]==prev[4])
                                {
                                    reserved[row[3]][row[4]]=true;
                                    row[5]=k; // Save the index to the previous column
                                    found=true;
                                    break;
                                }
                            }
                        }
                        
                        if(!found)
                        {
                            arr[i].remove(j);
                            j--;
                        }
                    }
                    
                }
                
// Keeping this here for a moment
//                // At this point we can remove duplicate ink/paper output combos too
//                boolean reserved[][]=new boolean[8][8];
//                
//                for(int j=0;j<arr[i].size();j++) // All rows again
//                {
//                    int row[]=arr[i].get(j);
//                    
//                    if(reserved[row[3]][row[4]]) // A-ha: someone has the same output combo
//                    {
//                        arr[i].remove(j);
//                        j--;
//                    }
//                    else
//                        reserved[row[3]][row[4]]=true;
//                }
                
                // No go
                if(arr[i].size()==0 || blocks[i].length>2)
                {   
                    errorfound=true;
                    
                    if(!fix)
                    {
                        setpixels(i*blox,y, (40-i)*blox,1, errorcolor);
                        break;
                    }
                    
                    if(fix && arr[i].size()==0) // No possible combinations found
                    {
                        // Find something... it's kind of vain if we get to this point
                        if(i==0)
                        {
                            addblock(arr[i], 7,0, 0x10+blocks[i][0], 7,blocks[i][0]);
                            setpixels(i*blox,y, blox,1, rgb[blocks[i][0]]);
                        }
                        else
                        {
                            int prev[]=arr[i-1].get(0);
                            addblock(arr[i], prev[3],prev[4], 0x10+blocks[i][0], prev[3],blocks[i][0]);
                            arr[i].get(0)[5]=0;
                            setpixels(i*blox,y, blox,1, rgb[blocks[i][0]]);
                        }
                    }
                }
            }
            
            if(!errorfound || fix)
            {
                // Walk from right to left and construct one of the possible paths
                int previdx=0;
                for(int i=39;i>=0;i--)
                {
                    int row[]=arr[i].get(previdx);
                    screen[y*(xsize/blox)+i]=row[2]; // operation
                    previdx=row[5]; // Next in line
                    
                    //println(str(row[0])+" "+str(row[1])+" "+str(row[2])+" "+str(row[3])+" "+str(row[4]));
                }
            }
        }

        pic.updatePixels();
        
        if(errorfound)
        {
            message("Color clash(es) found");
            return false;
        }
        else
        {
            message("Colors ok");
            return true;
        }
    }
    
    // Add an in/out combination to the list
    void addblock(ArrayList<int[]> arr,int inkin,int paperin,int operation,int inkout,int paperout)
    {
        arr.add(new int[] {inkin,paperin, operation, inkout,paperout, -1});
    }
    
    // Return the on bits of a block of pixels
    int bits(int pix[],int ink)
    {
        int b=0;
        
        for(int i=0;i<6;i++)
            if(pix[i]==ink)
                b|=0x20>>i;

        return b;
    }
    
    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
        load_hrs_b=new Button(x,super.addloadbuttons(x,y),"load hrs");
        return load_hrs_b.next;
    }
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_tap_b=new Button(x,super.addsavebuttons(x,y),"save tap");
        save_hrs_b=new Button(x,save_tap_b.next,"save hrs");
        save_asm_b=new Button(x,save_hrs_b.next,"save asm");
        return save_asm_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_asm_b.mouseover() || save_tap_b.mouseover() || save_hrs_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                    if(save_tap_b.mouseover()) // Tap
                    {
                        String s=fileselector(prefs.path,"TAP files",new String[] {"tap"},SAVE);
                        if(s!=null)
                            save_tap(s);
                    }
                    if(save_hrs_b.mouseover()) // HRS
                    {
                        String s=fileselector(prefs.path,"Hires files",new String[] {"hrs","hir"},SAVE);
                        if(s!=null)
                            save_hrs(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
        
        if(load_hrs_b.mouseover())
        {
            String s=fileselector(prefs.path,"Hires files",new String[] {"hrs","hir"},LOAD);
            if(s!=null)
                load_hrs(s);
        }
    }
    
    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        PrintWriter f=createWriter(name);
        
        f.println("\t.global\t_bitmap,_text");
        f.println();
        f.print("_bitmap:");
        printarr(f,screen,".byte","$");
        
        f.println();
        f.print("_text:");
        int text[]=new int[40*3];
        for(int i=0;i<text.length;i++)
            text[i]=0x10+border;
        printarr(f,text,".byte","$");
        
        f.println();

        f.flush();
        f.close();
        message("Saved: "+name);
    }
    
    // Save TAP
    void save_tap(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/oric-hires.tap");
        
        for(int i=0;i<screen.length;i++)
            b[233+i]=(byte)screen[i];
        for(int i=0;i<40*3;i++)
            b[8233+i]=(byte)(0x10+border);
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Load HRS/HIR
    void load_hrs(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<8000)
        {
            message("Can't open file: "+name);
            return;
        }

        message("Loaded: "+name);
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        int off=b.length-8000;
        
        // Convert the weird bitmap back to pixels
        for(int y=0;y<ysize;y++)
        {
            int ink=7,
                paper=0;
            
            for(int x=0;x<xsize;x+=blox,off++)
            {
                int c=(int)b[off],
                    xorri=((c&0x80)==0)? 0:0xffffff;
                    
                c&=0x7f;
                
                if(c<8) // Change ink - show paper
                {
                    ink=c;
                    for(int i=0;i<blox;i++)
                        sourcepic.pixels[y*xsize+x+i]=rgb[paper]^xorri;
                    continue;
                }
                if(c>=16 && c<24) // Change and show paper
                {
                    paper=c-16;
                    for(int i=0;i<blox;i++)
                        sourcepic.pixels[y*xsize+x+i]=rgb[paper]^xorri;                            
                    continue;
                }
                if((c&0x40)!=0) // Bitmap
                {
                    for(int i=0;i<blox;i++)
                        if((c&(0x20>>i))==0)
                            sourcepic.pixels[y*xsize+x+i]=rgb[paper]^xorri;
                        else
                            sourcepic.pixels[y*xsize+x+i]=rgb[ink]^xorri;
                    continue;
                }
                // Something else... let's show bg just in case
                for(int i=0;i<blox;i++)
                    sourcepic.pixels[y*xsize+x+i]=rgb[paper]^xorri;
            }
        }
        
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }
    // Save HRS/HIR
    void save_hrs(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/oric-hires.hrs");
        
        for(int i=0;i<screen.length;i++)
            b[23+i]=(byte)screen[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
}
