/*
    Plus/4 hires specific things
    
    Changelog:
    
    3.5.2016: Asm/prg exporters done and populate() fixed
    3.5.2016: The first version
*/

class Plus4Hires extends Machine
{
    // Grabbed this from the VICE screen:
    final int rgb[]={
        #000000,#2C2C2C,#621307,#00424C,#510378,#004E00,#27188E,#303E00,#582100,#463000,#244400,#630448,#004E0C,#0E2784,#33118E,#184800,
        #000000,#3B3B3B,#702419,#00505A,#601685,#125D00,#36289B,#3F4C00,#663100,#553F00,#345200,#711656,#005C1D,#1F3691,#42229B,#285700,
        #000000,#424242,#772C21,#055861,#661E8C,#1B6400,#3E30A2,#475400,#6D3900,#5C4700,#3B5900,#771F5D,#046325,#273E98,#492AA1,#305E00,
        #000000,#515151,#843B31,#17656F,#742E99,#2B7100,#4C3FAF,#556200,#7A4709,#6A5500,#4A6700,#852F6B,#177135,#364CA5,#5739AE,#3F6B00,
        #000000,#7A7A7A,#AC665C,#468E97,#9C5AC0,#57992E,#766AD5,#7E8A13,#A2713A,#927E20,#748F14,#AC5A93,#459960,#6276CB,#8064D4,#6A9419,
        #000000,#959595,#C58178,#62A8B1,#B675D9,#73B34C,#9185ED,#99A433,#BB8C57,#AC993E,#8FAA34,#C676AD,#62B37B,#7D91E4,#9B80ED,#85AE38,
        #000000,#AFAFAF,#DE9B93,#7DC2CA,#CF90F2,#8DCD68,#AB9FFF,#B3BE51,#D5A673,#C6B35B,#A9C351,#DF91C7,#7DCC96,#97ABFD,#B59AFF,#9FC755,
        #000000,#E1E1E1,#FFCFC6,#B2F4FC,#FFC4FF,#C1FE9D,#DDD2FF,#E5F088,#FFD9A8,#F7E591,#DBF588,#FFC4F9,#B1FEC9,#CBDDFF,#E7CDFF,#D2F98C
    };
    
    int luminance[];
                     
    Button save_prg_b,
           save_asm_b;
    
    Plus4Hires()
    {
        number=PLUS4_HIRES;
        name="Plus/4 Hires";
        
        xsize=320;
        ysize=200;
        blox=bloy=gridx=gridy=8;
        
        renderx=xsize*prefs.zoom;
        rendery=ysize*prefs.zoom;
        if(prefs.zoom>1)
        {
            switch(prefs.aspect)
            {
                case NTSC: rendery=ysize*prefs.zoom*5/4; break;
                default:   rendery=ysize*prefs.zoom; break;
            }
        }
        
        colorselheight=(prefs.UIWIDTH-2*prefs.BWIDTH)*2;
        
        border=110;
        maxborder=127;
        pen=113;
        errorcolor=#10bb10;
        gridcolor=#205050;
        
        super.rgb=rgb;
    }
    
    // Draw the color selector as columns
    void drawcolorsel(int px,int py)
    {
        int wide=(prefs.UIWIDTH-2*prefs.BWIDTH)/4,
            divisor=4,
            selectedx=0,
            selectedy=0,
            selected=0;
            
        if(rgb.length>16)
        {
            wide/=2;
            divisor=8;
        }
        
        for(int x=0,i=0;x<divisor;x++)
        {
            for(int y=0;y<rgb.length/divisor;y++,i++)
            {
                fill(rgb[i]);
                rect(px+x*wide,py+y*wide, wide,wide);
                
                if(mousePressed) // A bit of a hack, yes
                {
                    if(inside(px+x*wide,py+y*wide, px+x*wide+wide,py+y*wide+wide))
                        border=i;
                }

                if(i==border)
                {
                    selected=i;
                    selectedx=px+x*wide;
                    selectedy=py+y*wide;
                }
            }
        }
        
        noFill();
        stroke(rgb[selected]^0x808080);
        rect(selectedx,selectedy, wide,wide);
        noStroke();
        
        fill(prefs.UITEXT);
        text("border: "+str(border),px,py+wide*rgb.length/divisor+prefs.UIROW-7);
    }
    
    // Fill colormap and bitmap arrays (two-color 8x8 blocks here)
    void populate()
    {
        int bx=xsize/blox,
            by=ysize/bloy;
        
        colormap=new int[bx*by];
        luminance=new int[bx*by];
        bitmap=new int[bx*by*8];
        
        pic.loadPixels();
        for(int y=0,i=0;y<ysize;y+=bloy)
            for(int x=0;x<xsize;x+=blox,i++)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
                
                colormap[i]=(count[0][0]&0xf)<<4; // fg = most common color
                luminance[i]=count[0][0]>>4;
                
                if(count.length>1)
                {
                    colormap[i]+=count[1][0]&0xf;
                    luminance[i]+=count[1][0]&0xf0;
                }
                    
                for(int j=0;j<8;j++) // Create the bitmap
                {
                    int bits=0;
                    for(int k=0;k<8;k++)
                        if(pix[j*8+k]==count[0][1])
                            bits+=(0x80>>k);
                    
                    bitmap[i*8+j]=bits;
                }
            }
    }
    
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_prg_b=new Button(x,super.addsavebuttons(x,y),"save prg");
        save_asm_b=new Button(x,save_prg_b.next,"save asm");
        return save_asm_b.next;
    }
   
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_prg_b.mouseover() || save_asm_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                if(check(false)) // Seems to be fine
                {
                    if(save_prg_b.mouseover()) // PRG
                    {
                        String s=fileselector(prefs.path,"PRG files",new String[] {"prg"},SAVE);
                        if(s!=null)
                            save_prg(s);
                    }
                    if(save_asm_b.mouseover()) // Assembler
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                }
                else
                {
                    colorcorrect();
                    message("Image has color clash(es)");
                }
            }
        }
    }
    
    // Save PRG
    void save_prg(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/plus4-hires.prg");
        
        populate(); // Create the bitmap/colormap  
        
        // Replace some bytes
        b[0x2077]=(byte)border;
        for(int i=0;i<bitmap.length;i++)
            b[0x137+i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[0x2078+i]=(byte)colormap[i];
        for(int i=0;i<colormap.length;i++)
            b[0x2460+i]=(byte)luminance[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap
        
        PrintWriter f=createWriter(name);
        
        f.print("bitmap");
        printarr(f,bitmap,".byte","$");
        f.println();
        f.print("color");
        printarr(f,colormap,".byte","$");
        f.println();
        f.print("luma");
        printarr(f,luminance,".byte","$");
        f.println();
        f.println("border\t.byte\t$"+hex(border,2));

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
