/*
    Plus/4 multicolor specific things
    
    Changelog:

    18.5.2016: Smarter algorithm for bg color finding
    4.5.2016: Multi Botticelli saving/loading implemented
    4.5.2016: check() and populate() should work, exporters likewise
    3.5.2016: Removed most C-64 copypaste references
    3.5.2016: The first version
*/

class Plus4Multicolor extends Machine
{
    // Grabbed this from the VICE screen:
    final int rgb[]={
        #000000,#2C2C2C,#621307,#00424C,#510378,#004E00,#27188E,#303E00,#582100,#463000,#244400,#630448,#004E0C,#0E2784,#33118E,#184800,
        #000000,#3B3B3B,#702419,#00505A,#601685,#125D00,#36289B,#3F4C00,#663100,#553F00,#345200,#711656,#005C1D,#1F3691,#42229B,#285700,
        #000000,#424242,#772C21,#055861,#661E8C,#1B6400,#3E30A2,#475400,#6D3900,#5C4700,#3B5900,#771F5D,#046325,#273E98,#492AA1,#305E00,
        #000000,#515151,#843B31,#17656F,#742E99,#2B7100,#4C3FAF,#556200,#7A4709,#6A5500,#4A6700,#852F6B,#177135,#364CA5,#5739AE,#3F6B00,
        #000000,#7A7A7A,#AC665C,#468E97,#9C5AC0,#57992E,#766AD5,#7E8A13,#A2713A,#927E20,#748F14,#AC5A93,#459960,#6276CB,#8064D4,#6A9419,
        #000000,#959595,#C58178,#62A8B1,#B675D9,#73B34C,#9185ED,#99A433,#BB8C57,#AC993E,#8FAA34,#C676AD,#62B37B,#7D91E4,#9B80ED,#85AE38,
        #000000,#AFAFAF,#DE9B93,#7DC2CA,#CF90F2,#8DCD68,#AB9FFF,#B3BE51,#D5A673,#C6B35B,#A9C351,#DF91C7,#7DCC96,#97ABFD,#B59AFF,#9FC755,
        #000000,#E1E1E1,#FFCFC6,#B2F4FC,#FFC4FF,#C1FE9D,#DDD2FF,#E5F088,#FFD9A8,#F7E591,#DBF588,#FFC4F9,#B1FEC9,#CBDDFF,#E7CDFF,#D2F98C
    };
    
    int luminance[];
       
    int bg1,bg2;
        
    boolean forcedbg;
                         
    Button save_prg_b,
           save_asm_b,
           load_bot_b,
           save_bot_b;

    Checkbutton fix_bg_b;
    
    Plus4Multicolor()
    {
        number=PLUS4_MULTICOLOR;
        name="Plus/4 Multicolor";
        
        xsize=160;
        ysize=200;
        blox=gridx=4;
        bloy=gridy=8;
        
        renderx=xsize*prefs.zoom*2; // Pixels are about 2x as wide as high
        rendery=ysize*prefs.zoom;
        if(prefs.zoom>1)
        {
            switch(prefs.aspect)
            {
                case NTSC: rendery=ysize*prefs.zoom*5/4; break;
                default:   rendery=ysize*prefs.zoom; break;
            }
        }
        
        colorselheight=(prefs.UIWIDTH-2*prefs.BWIDTH)*2;
        
        border=110;
        maxborder=127;
        pen=113;
        bg1=bg2=0;
        errorcolor=#10bb10;
        gridcolor=#205050;
        
        super.rgb=rgb;
    }

    // Draw the color selector as columns
    void drawcolorsel(int px,int py)
    {
        int wide=(prefs.UIWIDTH-2*prefs.BWIDTH)/4,
            divisor=4,
            selectedx=0,
            selectedy=0,
            selected=0;
            
        if(rgb.length>16)
        {
            wide/=2;
            divisor=8;
        }
        
        for(int x=0,i=0;x<divisor;x++)
        {
            for(int y=0;y<rgb.length/divisor;y++,i++)
            {
                fill(rgb[i]);
                rect(px+x*wide,py+y*wide, wide,wide);
                
                if(mousePressed)
                {
                    if(mouseButton==LEFT)
                    {
                        if(inside(px+x*wide,py+y*wide, px+x*wide+wide,py+y*wide+wide))
                        {
                            if(!shift)
                                border=i;
                            else
                                bg1=i;
                        }
                    }
                    else
                        if(inside(px+x*wide,py+y*wide, px+x*wide+wide,py+y*wide+wide))
                            bg2=i;
                }
                
                if(i==border) // This needs to be drawn last
                {
                    selected=i;
                    selectedx=px+x*wide;
                    selectedy=py+y*wide;
                }
                
                if(i==bg1)
                {
                    fill(rgb[i]^0x808080);
                    triangle(px+x*wide,py+y*wide, px+x*wide+wide/2+1,py+y*wide, px+x*wide,py+y*wide+wide/2+1);
                }
                if(i==bg2)
                {
                    fill(rgb[i]^0x808080);
                    triangle(px+x*wide+wide,py+y*wide+wide, px+x*wide+wide/2,py+y*wide+wide, px+x*wide+wide,py+y*wide+wide/2);
                }
            }
        }
        
        noFill();
        stroke(rgb[selected]^0x808080);
        rect(selectedx,selectedy, wide,wide);
        noStroke();
        
        fill(prefs.UITEXT);
        text("border: "+str(border),px,py+wide*rgb.length/divisor+prefs.UIROW-7);
        text("b1: "+str(bg1)+" b2: "+str(bg2),px,py+wide*rgb.length/divisor+prefs.UIROW+13);
    }

    // Accept both 320x200 and 160x200
    boolean correctsize(int x,int y)
    {
        if((x==xsize || x==xsize*2) && y==ysize)
            return true;
        else
            return false;
    }
    
    // Check multicolor's 4-color limits
    boolean check(boolean fix)
    {
        boolean errorfound=false;
        
        if(pic==null)
        {
            message("Can't check an empty image");
            return false;
        }
        
        pic.loadPixels();
        
        // Find the two most common colors in >2 color blocks
        if(!forcedbg)
        {
            // Old simpler way
//            int kolors[][]=countcolors(pic.pixels,rgb);
//            bg1=kolors[0][0];
//            if(kolors.length>1)
//                bg2=kolors[1][0];
//            else
//                bg2=0;
        
            int freqs[]=new int[rgb.length];
            
            for(int y=0;y<ysize;y+=bloy)
                for(int x=0;x<xsize;x+=blox)
                {
                    int pix[]=getpixels(x,y,blox,bloy);
                    int count[][]=countcolors(pix,rgb);
                    
                    if(count.length>2) // Three or more color blocks
                    {
                        for(int i=0;i<count.length;i++) // Add frequency for each color
                            freqs[count[i][0]]++;
                    }                    
                }
                
            // Now we have freqs for each color in 2+ color blocks. Let's find the two most common.
            int largest=0;
            for(int i=0;i<freqs.length;i++)
                if(freqs[i]>largest)
                {
                    largest=freqs[i];
                    bg1=i;
                }
            largest=bg2=0;
            for(int i=0;i<freqs.length;i++)
                if(freqs[i]>largest && i!=bg1)
                {
                    largest=freqs[i];
                    bg2=i;
                }
        }
        int brgb1=rgb[bg1],
            brgb2=rgb[bg2];

        for(int y=0;y<ysize;y+=bloy) // Walk through every block
        {
            for(int x=0;x<xsize;x+=blox)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
 
                if(count.length>2) // More than two?
                {
                    boolean flagga=true;
                    if(count.length==3) // Three only... one needs to be the bg
                    {
                        for(int i=0;i<count.length;i++)
                        {
                            if(brgb1==count[i][1]) // Yes
                                flagga=false;
                            if(brgb2==count[i][1]) // Yes
                                flagga=false;
                        }
                    }
                    if(count.length==4) // Full four... both bgs needed
                    {
                        int found=0;
                        
                        for(int i=0;i<count.length;i++)
                        {
                            if(brgb1==count[i][1]) // Yes
                                found|=1;
                            if(brgb2==count[i][1]) // Yes
                                found|=2;
                        }
                        
                        if(found==3)
                            flagga=false;
                    }
                    
                    if(flagga) // More than two+bgs
                    {
                        errorfound=true;
                        if(fix)
                        {
                            if(count.length==3) // Special case: three colors, none is bg
                            {
                                int tmp[][]=new int[4][3];
                                for(int i=0;i<2;i++)
                                    arrayCopy(count[i],tmp[i]);
                                tmp[2][0]=bg1;
                                tmp[2][1]=brgb1;
                                tmp[3][0]=bg2;
                                tmp[3][1]=brgb2;
                                count=tmp;
                            }
                            
                            // Use the two most common colors + bgs for the remaining ones
                            int bi1=-1,bi2=-1;
                            
                            for(int i=0;i<4;i++)
                            {
                                if(count[i][1]==brgb1)
                                    bi1=i;
                                if(count[i][1]==brgb2)
                                    bi2=i;
                            }
                            
                            if(bi1==-1) // Not found
                            {
                                if(bi2!=2)
                                    bi1=2;
                                else
                                    bi1=3;
                            }
                            if(bi2==-1) // Not found
                            {
                                if(bi1!=3)
                                    bi2=3;
                                else
                                    bi2=2;
                            }
                                    
                            count[bi1][0]=bg1;
                            count[bi1][1]=brgb1;
                            count[bi2][0]=bg2;
                            count[bi2][1]=brgb2;
                            
                            for(int i=0;i<pix.length;i++)
                            {
                                int dist=1000000,best=0;
                                for(int j=0;j<4;j++)
                                {
                                    if(rgbdist(pix[i],count[j][1])<dist)
                                    {
                                        dist=rgbdist(pix[i],count[j][1]);
                                        best=count[j][1];
                                    }
                                }
                                pix[i]=best;
                            }
                            setpixels(x,y,blox,bloy,pix);
                        }
                        else
                            setpixels(x,y,blox,bloy,errorcolor);
                    }
                }
            }
        }
        pic.updatePixels();
        
        if(errorfound)
        {
            message("Color clash(es) found");
            return false;
        }
        else
        {
            message("Colors ok");
            return true;
        }
    }
    
    // Compile the bit- and colormaps
    void populate()
    {
        int off=0;
        
        bitmap=new int[40*25*8];
        colormap=new int[40*25];
        luminance=new int[40*25];
        
        pic.loadPixels();
        
        // Find the most common colors (need to rethink later)
        if(!forcedbg)
        {
            int kolors[][]=countcolors(pic.pixels,rgb);
            bg1=kolors[0][0];
            if(kolors.length>1)
                bg2=kolors[1][0];
            else
                bg2=0;
                
        }
        int brgb1=rgb[bg1],
            brgb2=rgb[bg2];

        for(int y=0;y<ysize;y+=bloy) // Walk through every block
        {
            for(int x=0;x<xsize;x+=blox)
            {
                int pix[]=getpixels(x,y,blox,bloy);
                int count[][]=countcolors(pix,rgb);
                int bloi=x/blox+40*(y/bloy);
                
                if(count.length>4)
                {
                    message("Can't save with color clashes.");
                    return;
                }

                // Form the color maps
                colormap[bloi]=luminance[bloi]=0;
                int cm1=ILLEGAL,cm2=ILLEGAL;
                
                for(int i=0,j=0;i<count.length;i++)
                {
                    if(count[i][1]!=brgb1 && count[i][1]!=brgb2)
                    {
                        if(j==0)
                        {
                            colormap[bloi]=count[i][0]<<4;
                            luminance[bloi]=count[i][0]>>4;
                            cm1=count[i][1];
                        }
                        if(j==1)
                        {
                            colormap[bloi]+=count[i][0]&0xf;
                            luminance[bloi]+=count[i][0]&0xf0;
                            cm2=count[i][1];
                        }
                        j++;
                    }
                }
                
                // Calculate the bitmap
                for(int i=0;i<bloy;i++,off++)
                {
                    bitmap[off]=0;
                    for(int j=0;j<4;j++) // Check which of the 4 choices fits best
                    {
                        int bits=0,
                            diff,
                            kolor=pix[j+i*4];
                        
                        diff=rgbdist(kolor,brgb1);
                        if(rgbdist(kolor,brgb2)<diff)
                        {
                            diff=rgbdist(kolor,brgb2);
                            bits=3;
                        }
                        if(rgbdist(kolor,cm1)<diff)
                        {
                            diff=rgbdist(kolor,cm1);
                            bits=1;
                        }
                        if(rgbdist(kolor,cm2)<diff)
                            bits=2;
                            
                        bitmap[off]+=((bits<<6)>>j*2);
                    }
                }
            }
        }
    }
    
    // Swap nibbles for Botticelli bg colors
    int swap(int a)
    {
        return ((a&0xf)<<4) + ((a&0xf0)>>4);
    }
    
    // Add more load buttons
    int addloadbuttons(int x,int y)
    {
        load_bot_b=new Button(x,super.addloadbuttons(x,y),"load botticel");
        return load_bot_b.next;
    }
    // Add more save buttons
    int addsavebuttons(int x,int y)
    {
        save_prg_b=new Button(x,super.addsavebuttons(x,y),"save prg");
        save_bot_b=new Button(x,save_prg_b.next,"save botticel");
        save_asm_b=new Button(x,save_bot_b.next,"save asm");
        return save_asm_b.next;
    }
    // Even an extra button
    int addextrabuttons(int x,int y)
    {
        fix_bg_b=new Checkbutton(x,super.addextrabuttons(x,y)+prefs.UIGAP,"fixed bg");
        return fix_bg_b.next;
    }
    
    // Handle our extra buttons
    void handleclicks()
    {
        super.handleclicks();
        
        if(save_prg_b.mouseover() || save_asm_b.mouseover() || save_bot_b.mouseover())
        {
            if(pic==null)
                message("Can't save and empty image");
            else
            {
                boolean forced=forcedbg; // Force it temporarily to keep the calculated fg/bg
                forcedbg=true;
                
                if(check(false)) // Seems to be fine
                {
                    if(save_prg_b.mouseover()) // PRG
                    {
                        String s=fileselector(prefs.path,"PRG files",new String[] {"prg"},SAVE);
                        if(s!=null)
                            save_prg(s);
                    }
                    if(save_bot_b.mouseover()) // Botticelli
                    {
                        String s=fileselector(prefs.path,"Multi Botticelli files",new String[] {"prg"},SAVE);
                        if(s!=null)
                            save_bot(s);
                    }
                    if(save_asm_b.mouseover()) // Asm
                    {
                        String s=fileselector(prefs.path,"Asm files",new String[] {"asm","s"},SAVE);
                        if(s!=null)
                            save_asm(s);
                    }
                }
                else
                {
                    //colorcorrect();
                    message("Image has color clash(es)");
                }
                forcedbg=forced;
            }
        }
        
        if(load_bot_b.mouseover())
        {
            String s=fileselector(prefs.path,"Multi Botticelli files",new String[] {"prg"},LOAD);
            if(s!=null)
                load_bot(s);
        }
    }
    
    // Save PRG
    void save_prg(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/plus4-multicolor.prg");
        
        populate(); // Create bitmap/colormap/luminance
        
        // Replace some bytes
        b[0x207e]=(byte)border;
        b[0x207f]=(byte)bg1;
        b[0x2080]=(byte)bg2;
        
        for(int i=0;i<bitmap.length;i++)
            b[0x13e+i]=(byte)bitmap[i];
        for(int i=0;i<colormap.length;i++)
            b[0x2081+i]=(byte)colormap[i];
        for(int i=0;i<luminance.length;i++)
            b[0x2469+i]=(byte)luminance[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    
    // Save asm
    void save_bot(String name)
    {
        if(pic==null)
            return;

        // Read template
        byte b[]=loadBytes("templates/plus4-m.botticelli.prg");
        
        populate(); // Create bitmap/colormap/luminance
        
        // Replace some bytes
        b[1025]=(byte)swap(bg1);
        b[1024]=(byte)swap(bg2);
        
        for(int i=0;i<bitmap.length;i++)
            b[2050+i]=(byte)bitmap[i];
        for(int i=0;i<luminance.length;i++)
            b[2+i]=(byte)luminance[i];
        for(int i=0;i<colormap.length;i++)
            b[1026+i]=(byte)colormap[i];
        
        saveBytes(name,b);        
        message("Saved: "+name);
    }
    // Load Botticelli
    void load_bot(String name)
    {
        byte b[]=loadBytes(name);
        if(b==null || b.length<10050)
        {
            message("Can't open file: "+name);
            return;
        }

        message("Loaded: "+name);
        prefs.filename=name;
        
        sourcepic=new PImage(xsize,ysize,RGB);
        sourcepic.loadPixels();
        
        bg1=swap(b[1025])&0x7f;
        bg2=swap(b[1024])&0x7f;
        
        for(int y=0,off=2050,boff=0;y<ysize;y+=bloy) // Blocks
            for(int x=0;x<xsize;x+=blox,boff++)
            {
                for(int i=0;i<bloy;i++,off++) // Block rows
                {
                    for(int j=0;j<blox;j++)
                    {
                        int bits=(((int)b[off])>>(6-j*2))&3, // Bitmap decides where the color comes from
                            kolor=bg1;
                            
                        if(bits==1)
                            kolor=((b[1026+boff]>>4)&0xf) + ((b[2+boff]&0x7)<<4); // Color+luma RAM
                        if(bits==2)
                            kolor=(b[1026+boff]&0xf) + ((b[2+boff])&0x70);      // Color+luma RAM
                        if(bits==3)
                            kolor=bg2;
                        
                        sourcepic.pixels[x+j+(y+i)*xsize]=rgb[kolor];
                    }
                }
            }
        
        sourcepic.updatePixels();
        if(pic==null)
            pic=new PImage(xsize,ysize,RGB);
        colorcorrect();
    }
    
    // Save asm
    void save_asm(String name)
    {
        if(pic==null)
            return;
        
        populate(); // Create the bitmap/colormap/screen ram
        
        PrintWriter f=createWriter(name);
        
        f.print("bitmap");
        printarr(f,bitmap,".byte","$");
        f.println();
        f.print("color");
        printarr(f,colormap,".byte","$");
        f.println();
        f.print("luma");
        printarr(f,luminance,".byte","$");
        f.println();
        f.println("bg1\t.byte\t$"+hex(bg1,2));
        f.println("bg2\t.byte\t$"+hex(bg2,2));
        f.println("border\t.byte\t$"+hex(border,2));

        f.flush();
        f.close();
        message("Saved: "+name);
    }
}
