/*
    Pixel Polizei - load, check and save pics for oldschool machines
    
    Changelog:
    
    18.5.2016: JR-200 added to the selector
    15.5.2016: Conditional repainting (boolean repaint) to save CPU time
    3.5.2016: A bit different wording for the menu to be cle
    2.5.2016: Switched to menuselector
    29.4.2016: Center the canvas if the UI column is taller
    26.4.2016: The first version
*/

Machine machine=null;

Preferences prefs;

PFont font;

// UI locations
int canvas_startx,canvas_starty,
    canvas_endx,canvas_endy,
    ui_startx,ui_startx2,ui_starty,
    lastbuttony;

boolean repaint=true;

void setup()
{
    frameRate(60);
    
    prefs=new Preferences();
    prefs.readprefs("prefs.txt");
    
    genericjavatheme();
    
    font=loadFont("arial.vlw");
    textFont(font);

    int result=selector("Pixel Polizei",
                        "C-64 | Hires,"+
                        "C-64 | Multicolor,"+
                        "CPC | Mode 0,"+
                        "CPC | Mode 0 (overscan),"+
                        "CPC | Mode 1,"+
                        "CPC | Mode 1 (overscan),"+
                        "MSX / Coleco | Screen 2,"+
                        "MSX | Screen 3,"+
                        "Oric | Hires,"+
                        "Panasonic JR-200,"+
                        "Plus/4 | Hires,"+
                        "Plus/4 | Multicolor,"+
                        "Sharp MZ-700,"+
                        "Sharp MZ-800 | Mode 1/2,"+
                        "ZX Spectrum"
                        );
                    
    delay(200); // Superstition?    
    switch(result)
    {
        case C64_HIRES:        machine=new C64Hires(); break;
        case C64_MULTICOLOR:   machine=new C64Multicolor(); break;
        case CPC_MODE0:        machine=new CPCMode0(false); break;        
        case CPC_MODE0OVER:    machine=new CPCMode0(true); break;
        case CPC_MODE1:        machine=new CPCMode1(false); break;
        case CPC_MODE1OVER:    machine=new CPCMode1(true); break;
        case MSX_SCREEN2:      machine=new MsxScreen2(); break;
        case MSX_SCREEN3:      machine=new MsxScreen3(); break;
        case ORIC:             machine=new Oric(); break;
        case PANASONIC_JR200:  machine=new PanasonicJR200(); break;
        case PLUS4_HIRES:      machine=new Plus4Hires(); break;
        case PLUS4_MULTICOLOR: machine=new Plus4Multicolor(); break;
        case SHARP_MZ700:      machine=new SharpMZ700(); break;
        case SHARP_MZ800:      machine=new SharpMZ800(); break;
        case ZX_SPECTRUM:      machine=new ZXSpectrum(); break;    
        default: System.exit(0); // Plain exit won't do
    }

    // Calculate the UI locations
    canvas_startx=prefs.BWIDTH;
    canvas_starty=prefs.BWIDTH;
    canvas_endx=canvas_startx+machine.renderx;
    canvas_endy=canvas_starty+machine.rendery;
    ui_startx=canvas_endx+prefs.BWIDTH;
    ui_startx2=ui_startx+prefs.BWIDTH;
    ui_starty=prefs.BWIDTH;
    
    // Add all the buttons
    int next;
    next=machine.addloadbuttons(ui_startx2,ui_starty);
    next=machine.addsavebuttons(ui_startx2,next+prefs.UIGAP);
    next=machine.addextrabuttons(ui_startx2,next+prefs.UIGAP);

    lastbuttony=next+prefs.UIGAP;
    
    size(ui_startx+prefs.UIWIDTH,
         max(canvas_endy+prefs.BWIDTH, lastbuttony+machine.colorselheight+prefs.UIROW+prefs.BWIDTH));
    
    if(canvas_endy+prefs.BWIDTH<height) // Y center canvas
    {
        canvas_starty=height/2-machine.rendery/2;
        canvas_endy=canvas_starty+machine.rendery;
    }
    
    noStroke();
    
    // Final settings
    if(prefs.grid)
    {
        machine.grid=true;
        machine.grid_b.checked=true;
    }
}

int oldx=-1,
    oldy=-1;

void draw()
{
    if(!focused)
    {
        shift=false;
    }
    
    // Check some conditions that need repainting the screen
    if(messagecounter+3*1000<millis() && messagecounter>0)
        repaint=true;

    if(mouseX!=oldx || mouseY!=oldy)
    {
        if(incanvas())
        {
            if(machine.zoom)
                repaint=true;
        }
        else
        {
            if(focused)
                repaint=true;
        }
    }

    machine.logic();

    if(repaint)
    {    
        frame.setTitle(machine.name+": "+prefs.filename);
        
        background(prefs.UICOLOR);
        fill(machine.rgb[machine.border]);
        rect(0,canvas_starty-prefs.BWIDTH, canvas_endx+prefs.BWIDTH,machine.rendery+2*prefs.BWIDTH);
        
        machine.drawcanvas();
        machine.drawcolorsel(ui_startx2,lastbuttony);
        
        drawbuttons();
        showmessages();
    }
    
    repaint=false;
    
    oldx=mouseX;
    oldy=mouseY;
}
