/*
    General-purpose functionality

    Changelog:
    
    19.5.2016: rgbdist() for arrays too
    3.5.2016: printarr() fix for sizes that aren't multiples of 8
    1.5.2016: printarr() for printing assembler data statements
    29.4.2016: Added countcolors() and moved bestmatch() here
    27.4.2016: Added timestamp()
    27.4.2016: Added ext() and getext() for dealing with filenames
    27.4.2016: Added copyimage()
    27.4.2016: Added rgbdist(), showmessages(), inside() and incanvas()
    26.4.2016: The first version
*/

import java.io.File;

int messagecounter=0;
String msg="";

void message(String text)
{
    messagecounter=millis();
    msg=text;
    //println(text);
    repaint=true;
}

void showmessages()
{
    // Needs to be visible from the background
    if(rgbdist(machine.rgb[machine.border],0)<300 || canvas_endy+prefs.BWIDTH<height-prefs.UIROW)
        fill(210);
    else
        fill(30);
    if(messagecounter+1000*3>millis())
        text(msg,prefs.BWIDTH,height-5);
    else
        messagecounter=0;
}

// Calculate RGB distance of two colors
int rgbdist(int a,int b)
{
    if(a==ILLEGAL || b==ILLEGAL)
        return(1000000);
    
    int dr=((a>>16)&0xff)-((b>>16)&0xff),
        dg=((a>>8)&0xff) -((b>>8)&0xff),
        db=((a>>0)&0xff) -((b>>0)&0xff);
        
    return (int)sqrt(dr*dr+dg*dg+db*db);
}

// ..or arrays 
int rgbdist(int a[],int b[])
{
    int error=0;
    
    for(int i=0;i<a.length && i<b.length;i++)
        error+=rgbdist(a[i],b[i]);
            
    return error;
}

// Count the frequencies of colors in an array
// Return an array with {colornum,color,count} on rows
int[][] countcolors(int pix[],int rgb[])
{
    int count[]=new int[rgb.length];
    
    // Count all
    for(int i=0;i<pix.length;i++)
    {
        count[bestmatch(pix[i],rgb)]++;
    }
       
    // How many different 
    int found=0;
    for(int i=0;i<count.length;i++)
        if(count[i]!=0)
            found++;
    
    // Create frequency table
    int result[][]=new int[found][3];
    for(int i=0,j=0;i<count.length;i++)
        if(count[i]!=0)
        {
            result[j][0]=i;
            result[j][1]=rgb[i];
            result[j][2]=count[i];
            j++;
        }
    
    // Better sort it too, highest count first
    boolean cont=true;
    while(cont)
    {
        int tmp[];
        cont=false;
        for(int i=0;i<result.length-1;i++)
        {
            if(result[i][2]<result[i+1][2])
            {
                cont=true;
                tmp=result[i];
                result[i]=result[i+1];
                result[i+1]=tmp;
            }
        }
    }
    
    return result;
}

// Find the best index for this RGB value
int bestmatch(int col,int rgb[])
{
    int diff=1000000,
        best=0;
    
    for(int i=0;i<rgb.length;i++)
    {
        if(rgb[i]!=ILLEGAL)
        {
            if(rgbdist(col,rgb[i])<diff)
            {
                best=i;
                diff=rgbdist(col,rgb[i]);
            }
        }
    }
    
    return best;
}
// And arrays too
int[] bestmatch(int col[],int rgb[])
{
    int tmp[]=new int[col.length];
    
    for(int i=0;i<col.length;i++)
        tmp[i]=bestmatch(col[i],rgb);
    
    return tmp;
}  

// Mouse inside this rect?
boolean inside(int left,int top,int right,int bottom)
{
    if(mouseX>=left && mouseY>=top && mouseX<=right && mouseY<=bottom)
        return true;
    else
        return false;
}

// Mouse inside the picture area?
boolean incanvas()
{
    return inside(canvas_startx,canvas_starty, canvas_endx,canvas_endy);
}

// Copy image pixels. Must be equal size!
void copyimage(PImage src,PImage dest)
{
    if(src==null || dest==null)
        return;
        
    src.loadPixels();
    dest.loadPixels();

    for(int i=0;i<src.width*src.height;i++)
        dest.pixels[i]=src.pixels[i];    

    dest.updatePixels();
}

// Change extension from something to something else
String ext(String name,String oldext,String newext)
{
    // Bad filename?
    if(name.length()<=oldext.length() || 
       (!name.substring(name.length()-oldext.length()).equalsIgnoreCase(oldext)))
    {
        message("Bad filename");
        return null;
    }
    
    return name.substring(0,name.length()-oldext.length())+newext;
}

// Return the extension or null if none
String getext(String name)
{
    if(name.length()<2)
        return null;
        
    for(int i=name.length()-1;i>0;i--)
        if(name.charAt(i)=='.')
            return name.substring(i);
    
    return null;
}

long timestamp(String name)
{
    File f=new File(name);
    return f.lastModified();
}

// Print an array of bytes in "assembly"
void printarr(PrintWriter f,int arr[],String db,String prefix)
{
    for(int i=0;i<arr.length;i+=8)
    {
        f.print("\t"+db+"\t");
        for(int j=0;j<8 && i+j<arr.length;j++)
        {
            f.print(prefix+hex(arr[i+j],2));
            if(j!=7 && i+j!=arr.length-1)
                f.print(",");
        }
        f.println();
    }
}
